/*
  Copyright 1996-2001
  Simon Whiteside

  $Id: TreeNodeObject.java,v 1.5 2001/05/09 10:30:18 sdw Exp $
*/
package simkin;
import java.util.*;

/**
 * This class provides and Executable interface to a TreeNode
 *<p>The method getValue, setValue and method all search for matching child labels within the TreeNode object. Only the first matching tag is used.  
 */
public class TreeNodeObject extends ExecutableRoot {
    /**
     * the underlying treenode itself
     */
    TreeNode m_Node;
    /**
     * This caches the parse trees for the methods already come across
     */
    Hashtable m_MethodCache=null;

    /**
     * Constructor 
     * @param node the treenode to be stored
     */
    public TreeNodeObject(TreeNode node){
	m_Node=node;
    }
    /**
     * @return the value of the treenode data as a string
     */
    public String toString(){
	return m_Node.data();
    }
    /**
     * @return true if the data in both nodes is the same
     */
    public boolean equals(Object o){
	return toString().equals(o.toString());
    }
    /**
     * @return the underlying treenode
     */
    public TreeNode getNode(){
	return m_Node;
    }
    /**
     * this method stores a value in the in-memory treenode object. If the value is of treenode type, the treenode is first copied and then stored
     * @param s the name of the value
     * @param v the value itself
     * @throws RuntimeException - if there was a problem running the script (such as not having permission to access a field)
     * @throws NoSuchFieldException - if the field could not be found
     */
    public void setValue(String s,String attribute,Object v) throws RuntimeException, NoSuchFieldException{
	TreeNode node_value=m_Node.findChild(s);
	if (node_value!=null){
	    if (v!=null){
		if (v instanceof TreeNodeObject)
		    node_value.copyItems(((TreeNodeObject)v).getNode());
		else
		    node_value.data(v.toString());
	    }
	}else
	    super.setValue(s,attribute,v);
    }
    /**
     * this method stores a value in the in-memory treenode object. If the value is of treenode type, the treenode is first copied and then stored
     * @param array_index the index of the item to update
     * @param v the value itself
     * @throws RuntimeException - if there was a problem running the script (such as not having permission to access a field)
     */
    public void setValueAt(Object array_index,String attribute,Object v) throws RuntimeException{
	TreeNode node_value=m_Node.nthChild(Interpreter.intValue(array_index));
	if (node_value!=null){
	    if (v!=null){
		if (v instanceof TreeNodeObject)
		    node_value.copyItems(((TreeNodeObject)v).getNode());
		else
		    node_value.data(v.toString());
	    }
	}else
	    super.setValueAt(array_index,attribute,v);
    }
    /**
     * this method retrieves a value from the in-memory treenode object. If the value retrieved is a treenode type, it is *not* copied
     * @param array_index the name of the value
     * @return the value or null
     * @throws RuntimeException - if there was a problem running the script (such as not having permission to access a field)
     */
    public Object getValueAt(Object array_index,String attribute) throws RuntimeException{
	Object ret=null;
	TreeNode node_value=m_Node.nthChild(Interpreter.intValue(array_index));
	if (node_value!=null){
	    ret=new TreeNodeObject(node_value);
	}else
	    ret=super.getValueAt(array_index,attribute);
	return ret;
    }
    /**
     * this method retrieves a value from the in-memory treenode object. If the value retrieved is a treenode type, it is *not* copied
     * @param s the name of the value
     * @return the value or null
     * @throws RuntimeException - if there was a problem running the script (such as not having permission to access a field)
     * @throws NoSuchFieldException - if the field could not be found
     */
    public Object getValue(String s,String attribute) throws RuntimeException, NoSuchFieldException{
	Object ret=null;
	if (s!=null && s.equals("numChildren"))
	    ret=new Integer(m_Node.numChildren());
	else{
	    TreeNode node_value=m_Node.findChild(s);
	    if (node_value!=null){
		ret=new TreeNodeObject(node_value);
	    }else
		ret=super.getValue(s,attribute);
	}
	return ret;
    }
    /**
     * This method returns an object which can iterate over all the immediate child node of this node
     */
    public TreeNodeObjectEnumerator enumerate(){
	return new TreeNodeObjectEnumerator(this);
    }
    /**
     * This method returns an object which can iterate over all the immediate child nodes of this node whose label matches the one given
     */
    public TreeNodeObjectEnumerator enumerate(String label){
	return new TreeNodeObjectEnumerator(this,label);
    }
    /**
     * This method writes the entire node tree to stdout
     */
    public void dump(){
	try{
	    m_Node.write(System.out,false);
	}catch(Exception e){
	    Tracer.trace(e);
	}
    }
    /**
     * this method attempts to find the named method within the in-memory treenode object. If it is found, the Interpreter is used to execute the method
     * @param s the name of the method
     * @param args an array of arguments to the method
     * @param ret the object to receive the result of the method call
     * @return true if the method was found, false otherwise
     * @throws ParseException if the code contained syntax errors
     * @throws NoSuchMethodException if the method could not be found
     * @throws RuntimeException if there was an error running the code
     */
    public Object method(String s,Object[] args) throws ParseException, RuntimeException,NoSuchMethodException{
	Object ret=null;
	if (m_Node!=null){
	    ParseNode parseNode=null;
	    if (m_MethodCache!=null)
		parseNode=(ParseNode)m_MethodCache.get(s);
	    if (parseNode==null){
		TreeNode node=m_Node.findChild(s);
		if (node!=null){
		    String code=node.data();
		    String location=Interpreter.addStrings(getClass().getName(),":",s);
		    ExecuteResult result=Interpreter.getInterpreter().executeString(location,this,code,args);
		    ret=result.getResult();
		    parseNode=result.getParseTree();
		    if (m_MethodCache==null)
			m_MethodCache=new Hashtable();
		    m_MethodCache.put(s,parseNode);
		}else
		    ret=super.method(s,args);
	    }else{	
		String location=Interpreter.addStrings(getClass().getName(),":",s);
		ret=Interpreter.getInterpreter().executeParseTree(location,this,parseNode,args);
	    }
	}else
	    ret=super.method(s,args);
	return ret;
    }
}
